﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Office.Server.ActivityFeed;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Utilities;
using Microsoft.Office.Server.UserProfiles;
using Microsoft.SharePoint.Administration.Claims;
using System.Text.RegularExpressions;

namespace SUGUK.Suggestions
{
    /// <summary>
    /// SUGUK Suggestions gatherer.
    /// </summary>
    class SuggestionsGatherer
    {
        #region Fields

        /// <summary>
        /// Field to hold the user profile manager.
        /// </summary>
        UserProfileManager userProfileManager = null;

        /// <summary>
        /// Field to hold the activity manager.
        /// </summary>
        ActivityManager activityManager = null;

        /// <summary>
        /// Collection to hold the items.
        /// </summary>
        List<ProcessingItem> items;

        #endregion

        #region Constructor

        /// <summary>
        /// Initializes a new instance of the <see cref="SuggestionsGatherer"/> class.
        /// </summary>
        public SuggestionsGatherer()
        {
            items = new List<ProcessingItem>();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SuggestionsGatherer"/> class.
        /// </summary>
        /// <param name="siteId">The site id.</param>
        public SuggestionsGatherer(Guid siteId)
        {
            items = new List<ProcessingItem>();

            ElevateManagers(siteId);
        }

        #endregion

        #region Activity Application

        /// <summary>
        /// Creates the activity application exists if it doesn't it creates it.
        /// </summary>
        /// <returns>The activity application.</returns>
        public ActivityApplication EnsureActivityApplication()
        {
            ActivityApplication app = null;

            if (SPServiceContext.Current != null)
            {
                ActivityManager activityManager = new ActivityManager();

                //Check to see that the current user has permissions to access the social data
                if (activityManager.PrepareToAllowSchemaChanges())
                {
                    app = activityManager.ActivityApplications[SuggestionsActivityFeedConstants.ActivityApplication];

                    if (app == null)
                    {
                        app = activityManager.ActivityApplications.Create(SuggestionsActivityFeedConstants.ActivityApplication);
                        app.Commit();
                        app.Refresh(true);
                    }

                    this.CreateActivityTypes(activityManager, app);

                    return app;
                }
                else
                {
                    try
                    {
                        //The process user doesn't have access to profiles service as manage social data role
                        return activityManager.ActivityApplications[SuggestionsActivityFeedConstants.ActivityApplication];
                    }
                    catch (ActivityFeedPermissionDeniedException ex)
                    {
                        //Just adding some extra info into the exception to help debugging
                        throw new ActivityFeedPermissionDeniedException("Must be User Profile Administrator to perform this operation. This is a defect in SP2010 where the user must have 'Full Control' on the profile service application. Add them with this role and run again", ex);
                    }
                }
            }

            return app;
        }

        /// <summary>
        /// Removes the activity application.
        /// </summary>
        /// <returns></returns>
        public bool RemoveActivityApplication()
        {
            if (SPServiceContext.Current != null)
            {
                ActivityManager activityManager = new ActivityManager();

                //Check to see that the current user has permissions to access the social data
                if (activityManager.PrepareToAllowSchemaChanges())
                {
                    ActivityApplication app = activityManager.ActivityApplications[SuggestionsActivityFeedConstants.ActivityApplication];
                    if (app != null)
                    {
                        try
                        {
                            //Before removing the application check it has:
                            // No activity event types > throw an exception if it does
                            // No activity event templates > throw an exception if it does

                            this.RemoveActivityTypes(activityManager, app);

                            app.Remove();
                        }
                        catch (NotImplementedException exNotImplemented)
                        {
                            //Do nothing here but log that it wasnt actioned as this is a SP2010 RTM bug
                            //do a log to operations
                        }
                    }
                }
            }

            return true;
        }

        #endregion

        #region Register/UnRegister Activity Types

        /// <summary>
        /// Creates the activities and templates.
        /// </summary>
        public void CreateActivityTypes(ActivityManager activityManager, ActivityApplication app)
        {
            using (SPMonitoredScope scope = new SPMonitoredScope("Executing CreateActivityTypes"))
            {
                if (app != null)
                {
                    ActivityType createType = app.ActivityTypes[SuggestionsActivityFeedConstants.ActivityType.Suggestion.ItemAdded];
                    if (createType == null)
                    {
                        createType = app.ActivityTypes.Create(SuggestionsActivityFeedConstants.ActivityType.Suggestion.ItemAdded);
                        createType.ActivityTypeNameLocStringResourceFile = SuggestionsActivityFeedConstants.ResourceFile;
                        createType.ActivityTypeNameLocStringName = SuggestionsActivityFeedConstants.ActivityType.Suggestion.ItemAddedDisplay;
                        createType.IsPublished = true;
                        createType.IsConsolidated = true;
                        createType.Commit();
                        createType.Refresh(true);
                    }

                    //Create the associated Single Value template
                    ActivityTemplate createTemplateSV = createType.ActivityTemplates[ActivityTemplatesCollection.CreateKey(false)];
                    if (createTemplateSV == null)
                    {
                        createTemplateSV = createType.ActivityTemplates.Create(false);
                        createTemplateSV.TitleFormatLocStringResourceFile = SuggestionsActivityFeedConstants.ResourceFile;
                        createTemplateSV.TitleFormatLocStringName = SuggestionsActivityFeedConstants.ActivityType.Suggestion.ItemAddedSVTemplate;
                        createTemplateSV.IsMultivalued = false;
                        createTemplateSV.Commit();
                        createTemplateSV.Refresh(true);
                    }
                }
            }
        }

        /// <summary>
        /// Removes the activities and templates.
        /// </summary>
        public void RemoveActivityTypes(ActivityManager activityManager, ActivityApplication app)
        {
            using (SPMonitoredScope scope = new SPMonitoredScope("Executing AnnouncementContentProcessor RemoveActivityTypes"))
            {
                if (app != null)
                {
                    ActivityType activityType = app.ActivityTypes[SuggestionsActivityFeedConstants.ActivityType.Suggestion.ItemAdded];
                    foreach (ActivityTemplate template in activityType.ActivityTemplates)
                    {
                        try
                        {
                            template.Remove();
                        }
                        catch (Exception ex)
                        {
                            //do the logging here
                            throw;
                        }
                    }

                    try
                    {
                        activityType.Remove();
                    }
                    catch (NotImplementedException ex)
                    {
                        //Do nothing here but log that it wasnt actioned as this is a SP2010 RTM bug
                        //do a log to operations
                    }
                }
            }
        }

        #endregion

        #region Create Activity Events
        
        /// <summary>
        /// Adds the list item for processing.
        /// </summary>
        /// <param name="currentItem">The current item.</param>
        public virtual void AddListItemForProcessing(SPListItem currentItem)
        {
            ProcessingItem pi = CreateProcessingItemFromListItem(currentItem);

            if(pi != null)
            {
                items.Add(pi);
            }
        }

        /// <summary>
        /// Creates the activity events.
        /// </summary>
        public virtual void CreateActivityEvents()
        {
            //Only bother to loop in if the containers exist
            if (this.items.Count > 0)
            {
                #region Create Activity Events

                foreach (ProcessingItem item in this.items)
                {
                    item.ActivityEvent = CreateAddedItemSingleValueActivityEvent(activityManager, item);
                }

                MulticastPublishedEvents(this.items, this.activityManager);

                #endregion
            }
        }

        /// <summary>
        /// Creates the processing item from list item.
        /// </summary>
        /// <param name="currentItem">The current item.</param>
        /// <returns></returns>
        protected virtual ProcessingItem CreateProcessingItemFromListItem(
            SPListItem currentItem)
        {
            ProcessingItem newItem = new ProcessingItem();

            SPFieldUserValue author;

            #region Check Moderation status

            if (currentItem.ParentList.EnableModeration)
            {
                //Reject any item that is not in a published state
                if (currentItem.ModerationInformation != null)
                {
                    if (currentItem.ModerationInformation.Status != SPModerationStatusType.Approved)
                    {
                        return null;
                    }
                }
            }

            if (currentItem.Level != SPFileLevel.Published)
            {
                return null;
            }

            #endregion

            #region Author

            //We process the author field first as we need to fail fast and hard if the user is not a real profile
            //For example the site system account authored the item
            if (currentItem.Fields.Contains(SPBuiltInFieldId.Author))
            {
                if (currentItem[SPBuiltInFieldId.Author] != null)
                {
                    try
                    {
                        author = new SPFieldUserValue(currentItem.Web, currentItem[SPBuiltInFieldId.Author].ToString());

                        if (author == null)
                        {
                            //The author should never be blank or not a SPFieldUserValue
                            throw new SPException("Unexpected condition, the Author field could not be cast to SPFieldUserValue for item (SiteId|WebId|ListId|ItemUniqueId): " + currentItem.Web.Site.ID.ToString() + "|" + currentItem.Web.ID.ToString() + "|" + currentItem.ParentList.ID.ToString() + "|" + currentItem.UniqueId.ToString());
                        }

                        if (author.User.LoginName == currentItem.Web.Site.SystemAccount.LoginName)
                        {
                            throw new SPException("Expected condition, the user was the System Account for item (SiteId|WebId|ListId|ItemUniqueId): " + currentItem.Web.Site.ID.ToString() + "|" + currentItem.Web.ID.ToString() + "|" + currentItem.ParentList.ID.ToString() + "|" + currentItem.UniqueId.ToString());
                        }

                        //currentItem.Web.Site.SystemAccount.
                        if (author.User.LoginName.Contains("|"))
                        {
                            SPClaimProviderManager mgr = SPClaimProviderManager.Local;
                            if (mgr != null)
                            {
                                newItem.Author = mgr.DecodeClaim(author.User.LoginName).Value;
                            }
                        }
                        else
                        {
                            newItem.Author = author.User.LoginName;
                        }
                    }
                    catch (FormatException)
                    {
                        newItem.Author = String.Empty;
                    }
                }
            }

            #endregion

            #region Suggestion Details

            Guid suggestionDetailsId = new Guid("{66aef3d9-580d-4252-a6d9-b2690e6179ba}");

            if (currentItem.Fields.Contains(suggestionDetailsId))
            {
                if (currentItem[suggestionDetailsId] != null)
                {
                    try
                    {
                        SPFieldMultiLineText bodyField = currentItem.Fields[suggestionDetailsId] as SPFieldMultiLineText;
                        if (bodyField != null)
                        {
                            //Ensure that any hyperlinks in the plain text get converted back to hyperlinks
                            newItem.PlainTextBodyWithHyperlinking = AddHyperlinkingToPlainTextField(bodyField, currentItem[suggestionDetailsId]);
                        }
                        else
                        {
                            newItem.PlainTextBodyWithHyperlinking = String.Empty;
                        }

                    }
                    catch (FormatException)
                    {
                        newItem.PlainTextBodyWithHyperlinking = String.Empty;
                    }
                }
            }

            #endregion

            #region Display Name

            if (currentItem.DisplayName != null)
            {
                newItem.DisplayName = currentItem.DisplayName;
            }

            #endregion

            #region Default View Url

            if (currentItem.ParentList != null)
            {
                if (currentItem.ParentList.ParentWeb != null)
                {
                    if (currentItem.ParentList.ParentWeb.Site != null)
                    {
                        newItem.ParentListDefaultViewUrl = currentItem.ParentList.ParentWeb.Site.MakeFullUrl(currentItem.ParentList.DefaultViewUrl);
                    }
                }
            }

            #endregion

            #region Default Display Form Url

            if (currentItem.ParentList != null)
            {
                if (currentItem.ParentList.ParentWeb != null)
                {
                    if (currentItem.ParentList.ParentWeb.Site != null)
                    {
                        newItem.ParentListDefaultDisplayFormUrl = currentItem.ParentList.ParentWeb.Site.MakeFullUrl(currentItem.ParentList.DefaultDisplayFormUrl);
                    }
                }
            }

            #endregion

            #region Id

            newItem.Id = currentItem.ID;

            #endregion

            #region Last Modified

            //Set the LastModified
            if (currentItem.Fields.Contains(SPBuiltInFieldId.Last_x0020_Modified))
            {
                if (currentItem[SPBuiltInFieldId.Last_x0020_Modified] != null)
                {
                    try
                    {
                        newItem.LastModified = Convert.ToDateTime(currentItem[SPBuiltInFieldId.Last_x0020_Modified].ToString());
                    }
                    catch (FormatException)
                    {
                        newItem.LastModified = DateTime.Now;
                    }
                }
            }

            #endregion

            #region List Id, Title

            if (currentItem.ParentList != null)
            {
                newItem.ListId = currentItem.ParentList.ID;
                newItem.ParentListTitle = currentItem.ParentList.Title;
            }

            #endregion

            #region Web Title, Id, Url, IsRootWeb

            if (currentItem.Web != null)
            {
                newItem.WebId = currentItem.Web.ID;
            }

            #endregion

            #region Site Id, Url

            if (currentItem.Web != null)
            {
                if (currentItem.Web.Site != null)
                {
                    newItem.SiteId = currentItem.Web.Site.ID;
                    
                }
            }

            #endregion

            #region Title

            try
            {
                if (currentItem.Fields.Contains(SPBuiltInFieldId.Title))
                {
                    if (currentItem.Title != null)
                    {
                        newItem.Title = currentItem.Title;
                    }
                }
            }
            catch (ArgumentException ex)
            {
                return null;
            }

            #endregion

            #region Get Entities

            if (!userProfileManager.UserExists(newItem.Author))
            {
                userProfileManager.CreateUserProfile(newItem.Author);
            }

            UserProfile currentUser = userProfileManager.GetUserProfile(newItem.Author);

            newItem.Owner = new MinimalPerson(currentUser).CreateEntity(activityManager);
            newItem.Publisher = new MinimalPerson(currentUser).CreateEntity(activityManager);

            Dictionary<long, MinimalPerson> owners;
            Dictionary<long, List<MinimalPerson>> colleaguesOfOwners = new Dictionary<long, List<MinimalPerson>>();
            List<MinimalPerson> interestedPeople = new List<MinimalPerson>();

            #region Get the colleagues
            
            List<long> publishers = new List<long>() { newItem.Publisher.Id };

            ActivityFeedGatherer.GetUsersColleaguesAndRights(activityManager, publishers, out owners, out colleaguesOfOwners);

            #endregion

            if (colleaguesOfOwners.ContainsKey(newItem.Publisher.Id))
            {
                //Add the colleagues to the interested people collection
                interestedPeople.AddRange(colleaguesOfOwners[newItem.Publisher.Id]);
            }

            newItem.InterestedPeople = interestedPeople;

            #endregion

            return newItem;
        }

        /// <summary>
        /// Adds the hyperlinking to plain text field.
        /// </summary>
        /// <param name="ItemBody">The item body.</param>
        /// <param name="currentItem">The current item.</param>
        /// <returns></returns>
        protected string AddHyperlinkingToPlainTextField(SPFieldMultiLineText ItemBody,
            object currentItem)
        {
            Regex regex = new Regex(@"\b(?:(?:https?|ftp|file)://|www\.|ftp\.)[-A-Z0-9+&@#/%=~_|$?!:,.]*[A-Z0-9+&@#/%=~_|$]\b", RegexOptions.IgnoreCase);
            string plainTextFieldValue = ItemBody.GetFieldValueAsText(currentItem);
            string croppedtext = string.Empty;
            MatchCollection matches = null;

            if (plainTextFieldValue.Length > 150)
            {
                croppedtext = plainTextFieldValue.Substring(0, 150);
                matches = regex.Matches(croppedtext);

                //If any urls are found in the first 150 characters then get the complete url if it is being cropped
                if (matches.Count > 0)
                {
                    //get the matches in the original body
                    MatchCollection originalmatches = regex.Matches(plainTextFieldValue);

                    //get the position of the last url in the cropped text within the actual body
                    int posoflasturl = plainTextFieldValue.IndexOf(originalmatches[matches.Count - 1].Value);

                    //calculate the end position of the url
                    int totallengthwithlasturl = posoflasturl + originalmatches[matches.Count - 1].Value.Length;

                    //if the url doesn not end within 150 characters 
                    if (totallengthwithlasturl > 150)
                    {
                        plainTextFieldValue = plainTextFieldValue.Substring(0, totallengthwithlasturl) + "...";
                    }
                }
                else
                {
                    plainTextFieldValue = croppedtext + "...";
                }
            }

            matches = regex.Matches(plainTextFieldValue);

            //Create the html anchor tags around the found value
            foreach (Match match in matches)
            {
                plainTextFieldValue = plainTextFieldValue.Replace(match.Value, "<a href='" + match.Value + "'>" + match.Value + "</a>");
            }

            return plainTextFieldValue;

        }
        
        /// <summary>
        /// Creates the single value announcement activity event.
        /// </summary>
        /// <param name="activityManager">The activity manager.</param>
        /// <param name="owner">The owner.</param>
        /// <param name="publisher">The publisher.</param>
        /// <param name="currentItem">The current item.</param>
        /// <returns>The activity event</returns>
        protected ActivityEvent CreateAddedItemSingleValueActivityEvent(ActivityManager activityManager,
            ProcessingItem currentItem)
        {
            using (SPMonitoredScope scope = new SPMonitoredScope("Executing CreateAddedItemSingleValueActivityEvent"))
            {
                ActivityType createType = activityManager.ActivityApplications[SuggestionsActivityFeedConstants.ActivityApplication].ActivityTypes[SuggestionsActivityFeedConstants.ActivityType.Suggestion.ItemAdded];
                if (createType != null)
                {
                    //Generate a new activity event
                    //Again if you haven't elevated the activity manager then this WILL throw access denied!!!
                    ActivityEvent activityEvent = ActivityEvent.CreateActivityEvent(activityManager, createType.ActivityTypeId, currentItem.Owner, currentItem.Publisher);

                    if (activityEvent != null)
                    {
                        activityEvent.Name = createType.ActivityTypeName;
                        activityEvent.ItemPrivacy = (int)Privacy.Public;
                        activityEvent.Owner = currentItem.Owner;
                        activityEvent.Publisher = currentItem.Publisher;
                        activityEvent.Value = SuggestionsActivityFeedConstants.ActivityType.Suggestion.SuggestionContentTypeValue;

                        DateTime itemLastModifiedUtc = currentItem.LastModified.ToUniversalTime();
                        activityEvent.DateValue = itemLastModifiedUtc;

                        activityEvent.PublishDate = itemLastModifiedUtc;
                        activityEvent.LastUpdateTime = itemLastModifiedUtc;

                        #region Web Link

                        //Create the Link to the suggestion list root link
                        Link suggestionList = new Link();

                        //Set the title to the list title
                        suggestionList.Name = currentItem.ParentListTitle;

                        //Set the href for the list default view
                        suggestionList.Href = SPHttpUtility.UrlPathEncode(currentItem.ParentListDefaultViewUrl, true);

                        //Set the link to this
                        activityEvent.Link = suggestionList;

                        #endregion

                        #region Suggestion Link

                        Link suggestionLink = new Link();

                        //Set the suggestionLink to the display name
                        if (!String.IsNullOrEmpty(currentItem.DisplayName))
                        {
                            suggestionLink.Name = currentItem.DisplayName;
                        }
                        else
                        {
                            suggestionLink.Name = currentItem.Title;
                        }

                        //Set the href of the Suggestion
                        suggestionLink.Href = SPHttpUtility.UrlPathEncode(currentItem.ParentListDefaultDisplayFormUrl + "?ID=" + currentItem.Id.ToString(), true);

                        suggestionLink.Value = currentItem.PlainTextBodyWithHyperlinking;

                        //Set the link 2
                        activityEvent.Link2 = suggestionLink;

                        #endregion

                        activityEvent.Commit();

                        return activityEvent;
                    }
                    return null;
                }
                return null;
            }
        }
        
        #endregion

        #region Publishing Events

        /// <summary>
        /// Multicasts the published events after generating the activity events for the interested colleagues.
        /// </summary>
        /// <param name="events">The events.</param>
        /// <param name="activityManager">The activity manager.</param>
        /// <remarks>By default this calls the ActivityFeedGatherer.GetUsersColleaguesAndRights to generate the interested colleagues.
        /// This will include people colleagues of the owner. To include or modify this process override the GetUsersAndInterestedColleagues method.</remarks>
        protected virtual void MulticastPublishedEvents(List<ProcessingItem> events,
            ActivityManager activityManager)
        {
            #region Guard

            if (events == null)
            {
                throw new ArgumentNullException("events is null.");
            }
            if (events.Count == 0)
            {
                return;
            }
            if (activityManager == null)
            {
                throw new ArgumentNullException("activityManager is null.");
            }

            #endregion

            foreach (ProcessingItem currentEvent in events)
            {
                Dictionary<long, List<ActivityEvent>> eventsPerOwner;

                Dictionary<long, List<MinimalPerson>> eventConsumersWithAccessPerPublisher = new Dictionary<long, List<MinimalPerson>>();

                eventConsumersWithAccessPerPublisher.Add(currentEvent.Publisher.Id, currentEvent.InterestedPeople);

                ActivityFeedGatherer.MulticastActivityEvents(activityManager, new List<ActivityEvent> { currentEvent.ActivityEvent }, eventConsumersWithAccessPerPublisher, out eventsPerOwner);

                List<ActivityEvent> eventsToMulticast;

                ActivityFeedGatherer.CollectActivityEventsToConsolidate(eventsPerOwner, out eventsToMulticast);

                WriteEvents(eventsToMulticast, activityManager);
            }
        }

        /// <summary>
        /// Writes the events.
        /// </summary>
        /// <param name="events">The events.</param>
        /// <param name="activityManager">The activity manager.</param>
        protected virtual void WriteEvents(List<ActivityEvent> events, ActivityManager activityManager) 
        { 
            int startIndex = 0; 

            while (startIndex + activityManager.MaxEventsPerBatch < events.Count) 
            { 
                ActivityFeedGatherer.BatchWriteActivityEvents(events, startIndex, activityManager.MaxEventsPerBatch); 
                startIndex += activityManager.MaxEventsPerBatch; 
            } 

            ActivityFeedGatherer.BatchWriteActivityEvents(events, startIndex, events.Count - startIndex); 
        }

        #endregion

        #region Managers

        /// <summary>
        /// Elevates the activity and profile manager and the subscription object.
        /// </summary>
        /// <param name="siteId">The site id.</param>
        protected virtual void ElevateManagers(Guid siteId)
        {
            #region Elevate the Managers

            SPSecurity.RunWithElevatedPrivileges(delegate()
            {
                using (SPSite aSite = new SPSite(siteId))
                {
                    try
                    {
                        //This is the creation of an elevated activity manager
                        //If you don't do this nothing works!!!
                        SPServiceContext context2 = SPServiceContext.GetContext(aSite);

                        userProfileManager = new UserProfileManager(context2);

                        string adminUserName = Environment.UserDomainName + "\\" + Environment.UserName;

                        if (!userProfileManager.UserExists(adminUserName))
                        {
                            userProfileManager.CreateUserProfile(adminUserName);
                        }

                        UserProfile p = userProfileManager.GetUserProfile(adminUserName);
                        activityManager = new ActivityManager(p, context2);
                    }
                    catch (UserProfileApplicationNotAvailableException ex)
                    {
                        //This is an example of logging to the operations version
                        //You might not need to throw... up to your needs
                        throw;
                    }

                }
            });

            #endregion
        }

        #endregion
    }
}
