﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.SharePoint.Administration;
using Microsoft.SharePoint;

namespace SUGUK.Suggestions.TimerJobs
{
    /// <summary>
    /// The SUGUK Suggestion timerjob which creates activity events from work items.
    /// </summary>
    public class SuggestionsActivityJobDefinition : SPWorkItemJobDefinition
    {
        #region Fields

        public static readonly Guid WorkItemTypeId = new Guid("{EB878884-F8FB-4E82-8E6B-5CEDC2FB755A}");

        SuggestionsGatherer gatherer;

        #endregion

        #region Properties

        /// <summary>
        /// Returns a globally unique identifier (GUID) that identifies the type of a specified work item (<see cref="T:Microsoft.SharePoint.SPWorkItem"/>) or work-item batch.
        /// </summary>
        /// <returns>
        /// Returns a GUID that specifies the type (<see cref="P:Microsoft.SharePoint.SPWorkItem.Type"/>) of a specified work item (<see cref="T:Microsoft.SharePoint.SPWorkItem"/>)
        /// </returns>
        public override Guid WorkItemType()
        {
            return WorkItemTypeId;
        }

        /// <summary>
        /// Gets an integer specifying the limit imposed on the batch size of the work-item collection.
        /// </summary>
        /// <value></value>
        /// <returns>Returns the <see cref="P:Microsoft.SharePoint.Administration.SPWebService.ContentService.WorkItemEventDeliveryBatchSize"/> (<see cref="P:Microsoft.SharePoint.Administration.SPWebService.ContentService"/>) for the work item collection.</returns>
        public override int BatchFetchLimit
        {
            get
            {
                return 50;
            }
        }

        /// <summary>
        /// Gets the display name.
        /// </summary>
        /// <value>The display name.</value>
        public override string DisplayName
        {
            get
            {
                return SuggestionsActivityFeedConstants.SuggestionsActivityFeedTimerJobTitle;
            }
        }    

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SuggestionsActivityJobDefinition"/> class.
        /// </summary>
        public SuggestionsActivityJobDefinition() 
        {     
        
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SuggestionsActivityJobDefinition"/> class.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="webApplication">The web application.</param>
        public SuggestionsActivityJobDefinition(string name, SPWebApplication webApplication)       
            : base(name, webApplication) 
        {     
        }

        #endregion

        #region Methods

        /// <summary>
        /// Processes the work items.
        /// </summary>
        /// <param name="contentDatabase">The content database.</param>
        /// <param name="workItems">The work items.</param>
        /// <param name="jobState">State of the job.</param>
        /// <returns></returns>
        protected override bool ProcessWorkItems(SPContentDatabase contentDatabase, SPWorkItemCollection workItems, SPJobState jobState)
        {
            if (workItems.Count > 0)
            {
                gatherer = new SuggestionsGatherer(workItems[0].SiteId);

                bool returnValue = base.ProcessWorkItems(contentDatabase, workItems, jobState);

                gatherer.CreateActivityEvents();

                return returnValue;
            }
            else
            {
                return base.ProcessWorkItems(contentDatabase, workItems, jobState);
            }
        }

        /// <summary>
        /// Processes the work item.
        /// </summary>
        /// <param name="contentDatabase">The content database.</param>
        /// <param name="workItems">The work items.</param>
        /// <param name="workItem">The work item.</param>
        /// <param name="jobState">State of the job.</param>
        /// <returns></returns>
        protected override bool ProcessWorkItem(SPContentDatabase contentDatabase, 
            SPWorkItemCollection workItems, SPWorkItem workItem, SPJobState jobState) 
        {        
            if (workItem == null) 
            {          
                throw new ArgumentNullException("workItem");     
            }         
            
            try 
            {          
                using (SPSite site = new SPSite(workItem.SiteId))
                {               
                    using (SPWeb web = site.OpenWeb(workItem.WebId)) 
                    {                   
                        try 
                        {         
                            SPList list = web.Lists[workItem.ParentId];       
                            SPListItem listItem = list.GetItemByUniqueId(workItem.ItemGuid);

                            gatherer.AddListItemForProcessing(listItem);

                        }               
                        catch (Exception ex)
                        {             
                            // exception handling     
                        }                
                        finally
                        {          
                            // SubCollection(): required to set the ParentWeb property on the WorkItemsCollection     
                            // Delete(): required to remove the item from the queue    
                            workItems.SubCollection(site, web, 0, (uint)workItems.Count).DeleteWorkItem(workItem.Id);    
                        }               
                    }      
                }     
            }     
            catch (Exception ex)
            {      
                // exception handling     
            }       
            return true;
        }

        #endregion
    }
}
